/*
 *    NzdlResponse.java
 *    Copyright (C) 2001 New Zealand Digital Library Project
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

//the package we're in
package org.nzdl.gsdl.service;

import java.util.HashMap;
import java.util.TreeMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;

import org.nzdl.gsdl.service.*;
import org.nzdl.gsdl.corba.gsdlInterface.*;
import org.nzdl.gsdl.util.*;

/**
 * An object that holds data from a response after a
 * {@link NzdlRequest NzdlRequest} has been serviced by
 * {@link NzdlService NzdlService}.
 *
 * @author Stuart Yeates  (say1@cs.waikato.ac.nz)
 * @author Aziz Mahoui    (amahoui@yahoo.com)
 * @author Gordon Paynter (paynter@cs.waikato.ac.nz)
 * @author Brett Sheeran (bsheeran@xtra.co.nz) (comments)
 * @version $Revision: 2688 $
 */

public class NzdlResponse extends java.lang.Object 
					implements java.io.Serializable{

  private corbaFilterResponseHolder m_responseHolder = null;
  private corbaFilterResponse m_response = null;
  private NzdlResultSet m_resultSet = null;
  private TreeMap m_frequencies = null;
  private List m_queryTerms = null;

  /**
   * Create an empty instance of a NzdlResponse object.
   */
  public NzdlResponse( ){

    m_responseHolder = NzdlCorbaFactory.createFilterResponseHolder();
    m_response = m_responseHolder.value;
  }

  /**
   * 
   */
  public void copyData(NzdlResponse old) {
    m_responseHolder = old.m_responseHolder ;
    m_response = old.m_response;
    m_resultSet =  old.m_resultSet;
    m_frequencies = old.m_frequencies;
    m_queryTerms = old.m_queryTerms;
  }

  /**
   * Get the corbaFilterReponse.
   */
  public corbaFilterResponse getResponse() {
    return m_response;
  }

  /**
   * Set the corbaFilterReponse.
   */
  public void setResponse(corbaFilterResponse response) {
    m_response = response;
    m_responseHolder.value = response;
    parseResponse();
  }

  /**
   * Get the corbaFilterReponse holder.
   */
  public corbaFilterResponseHolder getHolder() {
    return m_responseHolder;
  }

  /**
   * Get the result of {@link NzdlService NzdlService} servicing a
   * {@link NzdlRequest NzdlRequest}. <br>
   * Pre: <br>
   * - Created instances of: NzdlRequest, NzdlService and
   * {@link NzdlResponse NzdlResponse}. <br>
   * - NzdlService has serviced NzdlRequest. <br>
   * - Created instance of {@link NzdlResultSet NzdlResultSet}. <br>
   * Post: Copied result data from a NzdlResponse object into a NzdlResult
   * set object <br><br>
   *
   * <B>Example Code</B> <br>
   * <TT>NzdlQuery query = new NzdlQuery( "snail farming" );          <br>
   * NzdlRequest request = new NzdlRequest( query );                  <br>
   * NzdlResponse response = new NzdlResponse( );                     <br><br>
   * // use service to get response from request to "demo" collection <br>
   * NzdlService myService = new NzdlServiceClient( _args );          <br>
   * myService.service( "demo", request, response );                  <br><br>
   * // use getResultSet to get "results" from "response"             <br>
   * NzdlResultSet results = <B>response.getResultSet();</B> </TT>
   */
  public NzdlResultSet getResultSet() {
    if ( m_resultSet == null ) {
      parseResponse();
    }
    return m_resultSet;
  }

  /*
    This is a short expansion of the content of a corbaFilterResponse object.
    Some of the fields are not that clear to me (those with ?????). Therefore
    I am ignoring some of them for the time being.

    o Number of Documents
    o Type of results
    o Info on each of the search terms used:

      o Term string
      o Term frequency
      o Matched terms ?????

    o Info on each of the result docs returned:

      o Document ID
      o Result number ?????
      o Number of terms matched in this document
      o Number of phrases matched in this document
      o Array of document frequencies ?????
      o Info on the metadata:

         o Array of the names of the meta tags
         o Info on each of the meta data:

           o Params string ?????
           o Is the metadata a reference ?????
           o Array of values
	   o Id ?????
	   o Parentid ?????
  */


  /**
   * Parse the CORBA filter response object
   */
  private void parseResponse () {
    m_response = m_responseHolder.value;
    m_resultSet = new NzdlResultSet();
    m_resultSet.setNumOfDocs( m_response.numDocs );
    if (m_response.isApprox == null)
      throw new Error ("null m_response.isApprox in parseResponse");
    m_resultSet.setResultType( m_response.isApprox.value() );
    /** HAS SIDE EFFECT ! */
    parseTermsInfo( m_response.termInfo );
    /** HAS SIDE EFFECT ! */
    parseDocsInfo( m_response.docInfo );
  }

  private void parseTermsInfo( corbaTermInfo [] _termsInfo ) {
    String term = null;
    m_frequencies = new TreeMap();
    m_queryTerms = new ArrayList();
    for (int i=0; i<_termsInfo.length; i++) {
      term = NzdlCorbaFactory.toString(_termsInfo[i].term);
      m_queryTerms.add( term );
      m_frequencies.put( term, new Long(_termsInfo[i].frequency) );
    }
    m_resultSet.setQueryTerms( m_queryTerms );
    m_resultSet.setTermFrequencies( m_frequencies );
  }

  private void parseDocsInfo( corbaResultDocInfo [] _docsInfo ) {
    NzdlQueryHit queryHit = null;
    //System.err.println("Size of DocsInfo: " + _docsInfo.length);
    for (int i=0; i < _docsInfo.length; i++) {
      queryHit = new NzdlQueryHit();
      queryHit.setDocumentID( NzdlCorbaFactory.toString(_docsInfo[i].OID) );
      queryHit.setResultNumber( _docsInfo[i].resultNum );
      queryHit.setRanking( _docsInfo[i].ranking );
      queryHit.setNumOfTermsMatched( _docsInfo[i].termsMatched );
      queryHit.setNumOfPhrasesMatched( _docsInfo[i].phraseMatched );
      /** TO DO: extract corbaDocFreq_array value */
      queryHit.setMetaData( parseMetaData(_docsInfo[i].metadata) );
      // add to the result set
      m_resultSet.add( queryHit ) ;
    }
  }

  private Map parseMetaData( corbaMetadataInfo_map _metaData ) {
    /*
     * The key in the map will be the metaData Tag
     * The value for the key will be a Set of strings
     * Ex: key = "Author", value = {A. Mahoui, S. Franks}
     */
    Map metaData = new HashMap();
    // For each meta tag can be associated more than one value (example above)
    Set metaTagValues;
    corbaMetadataInfo c_metaDataInfo = null;

    for (int i=0; i < _metaData.names.length; i++) {
      // one meta data at a time please ...
      c_metaDataInfo = _metaData.values[i];

      // get a fresh meta tag values item
      metaTagValues = new HashSet();

      // get the set of values associated with this meta tag ...
      for (int j=0; j < c_metaDataInfo.values.length; j++) {
	metaTagValues.add( NzdlCorbaFactory.toString(c_metaDataInfo.values[j]) );
      }

      // map the meta tag with its set of values ...
      metaData.put( NzdlCorbaFactory.toString(_metaData.names[i]), metaTagValues);
    }
    return metaData;
  }

}

